<?php

if ( ! defined( 'ABSPATH' ) ) {
	// Exit if accessed directly
	exit;
}

/**
 * Class WC_TA_Taxamo_Manager
 *
 */
class WC_TA_Taxamo_Manager {

	const OIM_LINE_KEY = '_taxamo_line_key';

	public function setup() {

		// Add the Taxamo line key to ignored line item meta
		add_filter( 'woocommerce_hidden_order_itemmeta', array( $this, 'hide_line_key_from_meta_box' ), 10, 1 );

		// Store transaction
		add_action( 'woocommerce_checkout_update_order_meta', array( $this, 'store_transaction_on_checkout' ), 10, 1 );

		// Register payment
		add_action( 'woocommerce_payment_complete', array( $this, 'register_payment' ), 10, 1 );

		// Confirm transaction
		add_action( 'woocommerce_order_status_completed', array( $this, 'confirm_transaction' ), 10, 1 );
		
		add_action( 'woocommerce_order_status_processing', array( $this, 'confirm_transaction' ), 10, 1 );

		
		$setting_data = get_option('woocommerce_taxamo_settings');
		if (!empty($setting_data) ) {
			add_action( 'woocommerce_thankyou', array( $this, 'confirm_transaction' ), 10, 1 );
		}

		// Refund support
		add_action( 'woocommerce_order_refunded', array( $this, 'refund_transaction' ), 10, 2 );

		add_action( 'woocommerce_before_order_item_line_item_html', array( $this, 'migrate_tax_items' ), 10, 3 );

		add_filter( 'woocommerce_reports_taxes_rate', array( $this, 'modify_tax_by_code_report' ), 10, 3 );
	}

	/**
	 * Modifies the tax by code report so it shows correct percentage amount.
	 * This is needed because Taxamo generates tax rates dynamically and does not
	 * actually save a tax rate in the tax settings. The tax by code report pulls
	 * from the tax rate settings row and since it is not found, it will show blank.
	 *
	 * @since 1.2.11
	 * @version 1.2.11
	 * @return string
	 */
	public function modify_tax_by_code_report( $rate, $rate_id, $tax_row ) {
		$tax_rate = isset( $tax_row->tax_rate ) ? $tax_row->tax_rate : '';

		if ( ! empty( $tax_rate ) ) {
			$tax_rate_part = explode( '|', strtolower( $tax_rate ) );

			if ( preg_match( '/^vat.{4}e-service/', $tax_rate_part[0] ) ) {
				return $tax_rate_part[1];
			}

			if ( preg_match( '/^vat.{4}e-book/', $tax_rate_part[0] ) ) {
				return $tax_rate_part[1];
			}
		}

		return $rate;
	}

	/**
	 * Creates a hash rate id which is unique but repeatable.
	 *
	 * @todo Remove this code along with hook in the future perhaps WC3.2.
	 * @since 1.2.11
	 * @version 1.2.11
	 * @param string $rate_code
	 * @return string
	 */
	private function create_unique_rate_id( $rate_code ) {
		$rate_code_part = explode( '|', $rate_code );
		
		if (!isset($rate_code_part[0])) {
			$rate_code_part[0] = null;
		}

		if (!isset($rate_code_part[1])) {
			$rate_code_part[1] = null;
		}
		return (string) hexdec( substr( md5( strtolower( $rate_code_part[0] . '-' . $rate_code_part[1] ) ), 0, 15 ) );
	}

	/**
	 * Migrate pre WC30 tax rate ids to unqiue rate ids. This is done on the fly at run time.
	 * See https://github.com/woocommerce/woocommerce-taxamo/issues/98
	 *
	 * @todo Remove this code along with hook in the future perhaps WC3.2.
	 * @since 1.2.11
	 * @version 1.2.12
	 */
	public function migrate_tax_items( $item_id, $migrate_item, $order ) {
		if ( wc_tax_enabled() && version_compare( WC_VERSION, '3.0', '>=' ) ) {
			$order_taxes = $order->get_taxes();

			if (null != $migrate_item->get_taxes()) {
				$migrate_tax_data = $migrate_item->get_taxes();
				foreach ( $order_taxes as $migrate_tax_item ) {
					if ('yes' === WC_TA_Integration::$enable_v1) {
						$is_taxamo       = preg_match( '/taxamo/', array_keys( $migrate_tax_data['total'] )[0] );
					} else {
						$is_taxamo       = array_keys( $migrate_tax_data['total']);
					}
					$old_tax_rate_id = $migrate_tax_item->get_rate_id();

					//if ( $is_taxamo && ! empty( $migrate_item->get_meta( '_taxamo_line_key', true ) ) ) {
						$taxamo_line_key = $migrate_item->get_meta( '_taxamo_line_key', true );
					if ( $is_taxamo && ! empty($taxamo_line_key) ) {
						$old_tax_rate_label_arr = explode( ' ', $migrate_tax_item->get_label() );
						
						if ( ! isset( $old_tax_rate_label_arr[2] ) ) {
							continue;
						}

						$old_tax_rate_code = 'taxamo_' . strtolower( $old_tax_rate_label_arr[2] );
						$new_tax_rate_id   = $this->create_unique_rate_id( $migrate_tax_item->get_rate_code() );

						if ( ! empty( $migrate_tax_data['total'][ $old_tax_rate_code ] ) && ! empty( $migrate_tax_data['subtotal'][ $old_tax_rate_code ] ) ) {
							$migrate_tax_item_total    = isset( $migrate_tax_data['total'][ $old_tax_rate_code ] ) ? $migrate_tax_data['total'][ $old_tax_rate_code ] : '';
							$migrate_tax_item_subtotal = isset( $migrate_tax_data['subtotal'][ $old_tax_rate_code ] ) ? $migrate_tax_data['subtotal'][ $old_tax_rate_code ] : '';
							$migrate_tax_item->set_rate_id( $new_tax_rate_id );

							$migrate_item->set_taxes(
								array(
									'total'    => array( $new_tax_rate_id => $migrate_tax_item_total ),
									'subtotal' => array( $new_tax_rate_id => $migrate_tax_item_subtotal ),
								)
							);

							$migrate_item->save();
							$migrate_tax_item->save();
						}
					}
				}
			}
		}

		return;
	}

	/**
	 * Add the Taxamo line key to ignored line item meta
	 *
	 * @param $meta_keys
	 *
	 * @return array
	 */
	public function hide_line_key_from_meta_box( $meta_keys ) {
		$meta_keys[] = self::OIM_LINE_KEY;

		return $meta_keys;
	}

	/**
	 * Catch the new order action and update taxes to selected billing country
	 *
	 * @param $order_id
	 */
	public function store_transaction_on_checkout( $order_id ) {

		// Get the order
		$order = wc_get_order( $order_id );

		//$force_country = false;

		// Location confirmation
		//if ( isset( WC()->session ) && true === WC()->session->get( 'wc_ta_location_confirmation', false ) ) {
			// Set force country code
			$force_country = true;
		//}

		// Store transaciton
		$this->store_transaction( $order, $force_country );

	}

	public function set_EU_countriesss() {
		$countries = [
				'AT', 'BE', 'BG', 'CY',	'CZ', 'DE', 'DK', 'EE',	'ES', 'FI', 'FR', 'GR', 'HU', 'HR', 'IE', 'IT', 'LT', 'LU', 'LV', 'MT', 'NL', 'PL', 'PT', 'RO', 'SE', 'SI', 'SK'
			];

		return $countries;
	}

	/**
	 * Store an order as a transaction in Taxamo
	 *
	 * @param WC_Order $order
	 * @param bool $force_country
	 */
	public function store_transaction( $order, $force_country ) {

		$subtotal = $order->get_subtotal();
		if (isset($subtotal) && '150' <= $subtotal) {
			return true;
		}
		
		$payment_method = $order->get_payment_method();
		if ('cod' == $payment_method) {

			return true;
		}
		if (null != $order->get_shipping_address_1()) {

			$billing_country = $order->get_shipping_country();
		} else {

			$billing_country = $order->get_billing_country();
		}

		/* PLUGINS-2530 */

		$eu_country = $this->set_EU_countriesss();
		
		if (!in_array($billing_country, $eu_country)) {
			return true;
		}
		/* PLUGINS-2530 END */
		// The order manager
		$order_manager = new WC_TA_Order_Manager();

		// The transaction extra
		$transaction_extra = array();
		$invoice_address = array();

		// Check for VAT number
		$vat_number = opmc_hpos_get_post_meta( version_compare( WC_VERSION, '3.0', '<' ) ? $order->id : $order->get_id(), WC_TA_Vat_Number_Field::META_KEY, true );
		if ( '' !== $vat_number ) {
			// Set VAT number
			$transaction_extra['buyer_tax_number'] = $vat_number;
		}

		// Check for shipping/handling cost
		$order_shipping = version_compare( WC_VERSION, '3.0', '<' ) ? $order->order_shipping : $order->get_shipping_total();

		if ('yes' == WC_TA_Integration::$enable_v3) {

			//$transaction_extra['force_country_code'] = $billing_country;

			if (null != $order->get_shipping_address_1()) {

				$invoice_address = array(
					'street_one' => $order->get_shipping_address_1(),
					'street_two' => $order->get_shipping_address_2(),
					'city'        => $order->get_shipping_city(),
					'postal_code' => $order->get_shipping_postcode(),
					'country'     => $order->get_shipping_country(),
					);

			} else {

				$invoice_address = array(
					'street_one' => $order->get_billing_address_1(),
					'street_two' => $order->get_billing_address_2(),
					'city'        => $order->get_billing_city(),
					'postal_code' => $order->get_billing_postcode(),
					'country'     => $order->get_billing_country(),
				);

			}
			// Set invoice details in transaction_extra variable
			$invoice_address['invoice_address'] = $invoice_address;

		}

		if ( $order_shipping > 0 ) {
			/// Transaction lines
			$transaction_extra['transaction_lines'] = array();

			// Get the shipping line items
			$shipping_line_items = $order->get_items( 'shipping' );
			// Check if there are line items
			if ( count( $shipping_line_items ) > 0 ) {

				// Loop
				foreach ( $shipping_line_items as $shipping_line_item_key => $shipping_line_item ) {

					// Calculate some taxes
					$line_cost = $shipping_line_item['cost'];

					// Calculate shipping tax of this line item
					$shipping_tax_rate = 0;
					$order_shipping_tax = version_compare( WC_VERSION, '3.0', '<' ) ? $order->order_shipping_tax : $order->get_shipping_tax();
					$shipping_tax = 0;
					if ( $order_shipping_tax > 0 ) {
						$shipping_tax = $order_shipping_tax;
					}
					// Get store country name
					// Get store country name
					$store_country = WC()->countries->countries[WC()->countries->get_base_country()];
					$store_address     = get_option( 'woocommerce_store_address' );
					$store_address_2   = get_option( 'woocommerce_store_address_2' );
					$store_city        = get_option( 'woocommerce_store_city' );
					$store_postcode    = get_option( 'woocommerce_store_postcode' );

					$WC_Countries = new WC_Countries();
					$ccodes = $WC_Countries->get_countries();
					$country_code = array_search ($store_country, $ccodes);
					// Add shipping line item to transaction line
					$transaction_extra['transaction_lines'][] =
					 array(
						'product_class' => 'S',
						'custom_id'    => '' . $shipping_line_item_key,
						'quantity'     => 1,
						'amount' => wc_round_tax_total( $line_cost + $shipping_tax ),
						'description'  => $shipping_line_item['name'],
						'ship_from_address'	=>  array(
								'street_name' => $store_address,
								'address_detail' => $store_address_2,
								'city' => $store_city,
								'postal_code' => $store_postcode,
								'country_code' => $country_code
							),
								//'informative'  => true,
						);
				}
			} // End if().
		}

		// Setup request
		$request_store_transaction = new WC_TA_Request_Store_Transaction(
			$billing_country,
			$order_manager->get_items_from_order( $order ),
			$order->get_billing_first_name() . ' ' . $order->get_billing_last_name(), $order->get_billing_email(), $order->get_id(), $transaction_extra, $invoice_address
		);

		// Do request
		if ( $request_store_transaction->do_request() ) {

			// Get the body
			$response_body = $request_store_transaction->get_response_body();

			if ( isset( $response_body->transaction ) ) {

				/* PLUGINS-2530 */
				if ( isset( $response_body->transaction->buyer_tax_number_valid ) && true == $response_body->transaction->buyer_tax_number_valid ) {
					$valid_tax_number = true;
				
					opmc_hpos_update_post_meta( $order->get_id(), 'is_vat_exempt', 'yes');
				}
				/* PLUGINS-2530 */
				// Attach transaction to order
				opmc_hpos_update_post_meta( version_compare( WC_VERSION, '3.0', '<' ) ? $order->id : $order->get_id(), 'vertex_transaction_key', $response_body->transaction->key );

				// Loop through line items
				if ( count( $response_body->transaction->transaction_lines ) > 0 ) {

					foreach ( $response_body->transaction->transaction_lines as $transaction_line ) {

						if ( isset( $transaction_line->custom_id ) && isset( $transaction_line->line_key ) ) {

							// Add Taxamo line key to order item
							wc_add_order_item_meta( $transaction_line->custom_id, self::OIM_LINE_KEY, $transaction_line->line_key, true );

						}
					}
				}
			}
		} else {

			if ( function_exists( 'wc_add_notice' ) ) {
				wc_add_notice( $request_store_transaction->get_error_message(), 'error' );
			}
		} // End if().
	}

	/**
	 * Register the payment in Taxamo
	 *
	 * @param int $order_id
	 */
	public function register_payment( $order_id ) {

		// Get the order
		$order = wc_get_order( $order_id );

		// Get payment method title
		$payment_method_title = opmc_hpos_get_post_meta( $order_id, '_payment_method_title', true );

		// Get the Taxamo transaction key
		$transaction_key = opmc_hpos_get_post_meta( $order_id, 'vertex_transaction_key', true );

		$request_register_payment = new WC_TA_Request_Register_Payment( $transaction_key, $order->get_total(), $payment_method_title );

		// Do request
		/* if ( !empty($request_register_payment->do_request()) ) {
			/**
			 * e.g. trigger error.
			 *
			 * @todo Better error handling
			 *
		}*/

	}

	/**
	 * Confirm the transaction
	 *
	 * @param $order_id
	 */
	public function confirm_transaction( $order_id ) {

		// Get the order
		$order = wc_get_order( $order_id );
		// Get the Taxamo transaction key
		$transaction_key = opmc_hpos_get_post_meta( $order_id, 'vertex_transaction_key', true );
		// Check for enable_v3
		if ('yes' === WC_TA_Integration::$enable_v3) {

			if ( !empty($transaction_key) && 'processing' == $order->get_status() || 'completed' == $order->get_status()) {

				// Setup the Request
				$request_confirm_transaction = new WC_TA_Request_Confirm_Transaction( $transaction_key );
				// Do request
				$response = $request_confirm_transaction->do_request();
				$response_body = $request_confirm_transaction->get_response_body();
				
				if (!empty($response_body)) {
					$invoice_url = $response_body->transaction->transaction_lines[0]->invoice_image_url;
					$vertex_ioss_number = $response_body->transaction->transaction_lines[0]->ioss_number;
					
					opmc_hpos_add_post_meta( $order_id, 'vertex_ioss_number', $vertex_ioss_number );
					opmc_hpos_add_post_meta( $order_id, 'vertex_invoice_url', $invoice_url );
				}
			}

		} else {

			// Setup the Request
			$request_confirm_transaction = new WC_TA_Request_Confirm_Transaction( $transaction_key );

			// Do request
			$response = $request_confirm_transaction->do_request();
			$response_body = $request_confirm_transaction->get_response_body();

			if (!empty($response_body)) {

				if ('yes' === WC_TA_Integration::$enable_v1) {
					
					$invoice_url = esc_url( "https://dashboard.taxamo.com/api/v1/transactions/$transaction_key/invoice?public_token=" . WC_TA_Integration::$public_token );
					opmc_hpos_add_post_meta( $order_id, 'vertex_invoice_url', $invoice_url, true );
				}
				if ('yes' === WC_TA_Integration::$enable_v2) {

					$invoice_url = $response_body->transaction->invoice_image_url;
					opmc_hpos_add_post_meta( $order_id, 'vertex_invoice_url', $invoice_url, true );
				}
			}
		}
	}

	/**
	 * Refund a transaction
	 *
	 * @param $order_id
	 * @param $refund_id
	 *
	 * @return bool
	 */
	public function refund_transaction( $order_id, $refund_id ) {

		// Get the Taxamo transaction key
		$transaction_key = opmc_hpos_get_post_meta( $order_id, 'vertex_transaction_key', true );
		$order = wc_get_order( $order_id );

		$order_refunds = $order->get_refunds();

		// Get Refund reason
		$get_reason = '';
		foreach ( $order_refunds as $order_refund ) {
			// To be sure we check if that method exist and that is not empty
			if ( method_exists( $order_refund, 'get_reason' ) && $order_refund->get_reason() ) {
				$get_reason = $order_refund->get_reason();
			}
		} //

		// Get refund
		$refund = wc_get_order( $refund_id );

		// Get refund line items
		$refund_items = $refund->get_items( array( 'line_item', 'shipping' ) );

		// Check & loop
		
		if ( count( $refund_items ) > 0 ) {
			foreach ( $refund_items as $refund_item ) {

				// Get the line key
				$line_key = wc_get_order_item_meta( $refund_item['refunded_item_id'], self::OIM_LINE_KEY, true );
				
				// Only do the refund when the line item exists
				if ( '' !== $line_key ) {

					// Get correct amount
					$amount = abs( ( isset( $refund_item['line_total'] ) ? $refund_item['line_total'] : ( isset( $refund_item['cost'] ) ? $refund_item['cost'] : 0 ) ) );

					// Only continue if amount > 0
					if ( $amount > 0 ) {

						// Create refund request
						$refund_request = new WC_TA_Request_Refund( $transaction_key, $line_key, $amount, $get_reason );

						// Do the rquest
						$response = $refund_request->do_request();
					}
				}
			}
		}

		// Refund done
		return true;
	}

}

